Exception: HTTP2::Error::HandshakeError
- Defined in:
- lib/http/2/error.rb,
sig/error.rbs
Overview
Raised if connection header is missing or invalid indicating that this is an invalid HTTP 2.0 request - no frames are emitted and the connection must be aborted.