Class: HTTP2::Header::EncodingContext
- Inherits:
-
Object
- Object
- HTTP2::Header::EncodingContext
- Includes:
- Error
- Defined in:
- lib/http/2/header/encoding_context.rb,
sig/header/encoding_context.rbs
Constant Summary collapse
- UPPER =
/[[:upper:]]/.freeze
- STATIC_TABLE =
[ [":authority", ""], [":method", "GET"], [":method", "POST"], [":path", "/"], [":path", "/index.html"], [":scheme", "http"], [":scheme", "https"], [":status", "200"], [":status", "204"], [":status", "206"], [":status", "304"], [":status", "400"], [":status", "404"], [":status", "500"], ["accept-charset", ""], ["accept-encoding", "gzip, deflate"], ["accept-language", ""], ["accept-ranges", ""], ["accept", ""], ["access-control-allow-origin", ""], ["age", ""], ["allow", ""], ["authorization", ""], ["cache-control", ""], ["content-disposition", ""], ["content-encoding", ""], ["content-language", ""], ["content-length", ""], ["content-location", ""], ["content-range", ""], ["content-type", ""], ["cookie", ""], ["date", ""], ["etag", ""], ["expect", ""], ["expires", ""], ["from", ""], ["host", ""], ["if-match", ""], ["if-modified-since", ""], ["if-none-match", ""], ["if-range", ""], ["if-unmodified-since", ""], ["last-modified", ""], ["link", ""], ["location", ""], ["max-forwards", ""], ["proxy-authenticate", ""], ["proxy-authorization", ""], ["range", ""], ["referer", ""], ["refresh", ""], ["retry-after", ""], ["server", ""], ["set-cookie", ""], ["strict-transport-security", ""], ["transfer-encoding", ""], ["user-agent", ""], ["vary", ""], ["via", ""], ["www-authenticate", ""] ].each(&:freeze).freeze
- STATIC_TABLE_BY_FIELD =
STATIC_TABLE .each_with_object({}) .with_index { |((field, value), hs), idx| (hs[field] ||= []) << [idx, value].freeze } .each_value(&:freeze) .freeze
- STATIC_TABLE_SIZE =
STATIC_TABLE.size
- STATIC_ALL =
%i[all static].freeze
- STATIC_NEVER =
%i[never static].freeze
Instance Attribute Summary collapse
-
#current_table_size ⇒ Integer
readonly
Current table size in octets.
-
#settings ⇒ Settings
readonly
Current encoding settings.
-
#table ⇒ Array[header_pair]
readonly
Current table of header key-value pairs.
Instance Method Summary collapse
- #add_to_table ⇒ void
-
#addcmd(field, value) ⇒ header_command
Emits command for a +field+/+value+ header.
-
#dereference(index) ⇒ header_pair
Finds an entry in current dynamic table by
index. -
#dup ⇒ EncodingContext
Duplicates current compression context.
-
#encode(headers) {|arg0| ... } ⇒ void
Plan
headerscompression. -
#initialize(settings = Settings.new) ⇒ EncodingContext
constructor
Initializes compression context with appropriate client/server
settingsand maximum size of the dynamic table. - #listen_on_table { ... } ⇒ void
-
#process(cmd) ⇒ header_pair?
Header Block Processing - http://tools.ietf.org/html/draft-ietf-httpbis-header-compression-10#section-4.1.
- #resize_table(cmdsize) ⇒ void
-
#size_check?(cmdsize) ⇒ Boolean
whether
cmdfits in the dynamic table. -
#table_size=(size) ⇒ void
Alter dynamic table size.
Methods included from Error
Constructor Details
#initialize(settings = Settings.new) ⇒ EncodingContext
Initializes compression context with appropriate client/server
settings and maximum size of the dynamic table.
104 105 106 107 108 109 110 111 112 |
# File 'lib/http/2/header/encoding_context.rb', line 104 def initialize(settings = Settings.new) @table = [] @table_by_field = Hash.new { |hs, k| hs[k] = [] } @unshifts = 0 @settings = settings @limit = settings.table_size @_table_updated = false @current_table_size = 0 end |
Instance Attribute Details
#current_table_size ⇒ Integer (readonly)
Current table size in octets
100 101 102 |
# File 'lib/http/2/header/encoding_context.rb', line 100 def current_table_size @current_table_size end |
#settings ⇒ Settings (readonly)
Current encoding settings
97 98 99 |
# File 'lib/http/2/header/encoding_context.rb', line 97 def settings @settings end |
#table ⇒ Array[header_pair] (readonly)
Current table of header key-value pairs.
94 95 96 |
# File 'lib/http/2/header/encoding_context.rb', line 94 def table @table end |
Instance Method Details
#add_to_table ⇒ void
This method returns an undefined value.
49 |
# File 'sig/header/encoding_context.rbs', line 49
def add_to_table: (string name, string value) -> void
|
#addcmd(field, value) ⇒ header_command
Emits command for a +field+/+value+ header. Prefer static table over dynamic table. Prefer exact match over name-only match.
233 234 235 236 237 238 239 240 241 242 243 244 245 246 247 248 249 250 251 252 253 254 255 256 257 258 259 260 261 262 263 264 265 266 267 268 269 270 |
# File 'lib/http/2/header/encoding_context.rb', line 233 def addcmd(field, value) # @type var exact: Integer? exact = nil # @type var name_only: Integer? name_only = nil index_type = @settings.index if STATIC_ALL.include?(index_type) && STATIC_TABLE_BY_FIELD.key?(field) STATIC_TABLE_BY_FIELD[field].each do |i, svalue| name_only ||= i if value == svalue exact = i break end end end if index_type == :all && !exact field_entries = @table_by_field[field] field_entries&.each do |hvalue, unshift_id| abs_index = (@unshifts - unshift_id) + STATIC_TABLE_SIZE name_only ||= abs_index if value == hvalue exact = abs_index break end end end if exact { name: exact, type: :indexed } else { name: name_only || field, value: value, type: :incremental } end end |
#dereference(index) ⇒ header_pair
Finds an entry in current dynamic table by index.
Note that index is zero-based in this module.
If the index is greater than the last index in the static table,
an entry in the dynamic table is dereferenced.
If the index is greater than the last header index, an error is raised.
137 138 139 140 141 142 143 144 145 146 |
# File 'lib/http/2/header/encoding_context.rb', line 137 def dereference(index) # NOTE: index is zero-based in this module. return STATIC_TABLE[index] if index < STATIC_TABLE_SIZE idx = index - STATIC_TABLE_SIZE raise CompressionError, "Index too large" if idx >= @table.size @table[index - STATIC_TABLE_SIZE] end |
#dup ⇒ EncodingContext
Duplicates current compression context
115 116 117 118 119 120 121 122 123 124 125 126 127 128 |
# File 'lib/http/2/header/encoding_context.rb', line 115 def dup other = EncodingContext.new(@settings) t = @table tbf = @table_by_field.transform_values(&:dup) unshifts = @unshifts l = @limit other.instance_eval do @table = t.dup # shallow copy @table_by_field = tbf @unshifts = unshifts @limit = l end other end |
#encode(headers) {|arg0| ... } ⇒ void
This method returns an undefined value.
Plan headers compression.
213 214 215 216 217 218 219 220 221 222 223 224 225 226 227 |
# File 'lib/http/2/header/encoding_context.rb', line 213 def encode(headers) # Literals commands are marked with :noindex when index is not used noindex = STATIC_NEVER.include?(@settings.index) headers.each do |field, value| # Literal header names MUST be translated to lowercase before # encoding and transmission. field = field.downcase if UPPER.match?(field) value = "/" if field == ":path" && value.empty? cmd = addcmd(field, value) cmd[:type] = :noindex if noindex && cmd[:type] == :incremental process(cmd) yield cmd end end |
#listen_on_table { ... } ⇒ void
This method returns an undefined value.
279 280 281 282 283 |
# File 'lib/http/2/header/encoding_context.rb', line 279 def listen_on_table yield ensure @_table_updated = false end |
#process(cmd) ⇒ header_pair?
Header Block Processing
150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 195 196 197 198 199 200 201 202 203 204 205 206 207 208 209 210 |
# File 'lib/http/2/header/encoding_context.rb', line 150 def process(cmd) type = cmd[:type] name = cmd[:name] value = cmd[:value] case type when :changetablesize raise CompressionError, "tried to change table size after adding elements to table" if @_table_updated # we can receive multiple table size change commands inside a header frame. However, # we should blow up if we receive another frame where the new table size is bigger. table_size_updated = @limit != @settings.table_size raise CompressionError, "dynamic table size update exceed limit" if !table_size_updated && value > @limit self.table_size = value nil when :indexed # Indexed Representation # An _indexed representation_ entails the following actions: # o The header field corresponding to the referenced entry in either # the static table or dynamic table is added to the decoded header # list. dereference(name) when :incremental, :noindex, :neverindexed # A _literal representation_ that is _not added_ to the dynamic table # entails the following action: # o The header field is added to the decoded header list. # A _literal representation_ that is _added_ to the dynamic table # entails the following actions: # o The header field is added to the decoded header list. # o The header field is inserted at the beginning of the dynamic table. case name when Integer name, v = dereference(name) value ||= v when UPPER raise ProtocolError, "Invalid uppercase key: #{name}" end emit = [name, value] # add to table cmdsize = name.bytesize + value.bytesize + 32 if type == :incremental && size_check?(cmdsize) @table.unshift(emit) @unshifts += 1 @table_by_field[name].unshift([value, @unshifts]) @current_table_size += cmdsize @_table_updated = true end emit else raise CompressionError, "Invalid type: #{type}" end end |
#resize_table(cmdsize) ⇒ void
This method returns an undefined value.
287 288 289 290 291 292 293 294 295 296 297 298 299 300 301 |
# File 'lib/http/2/header/encoding_context.rb', line 287 def resize_table(cmdsize) return if @table.empty? while @current_table_size + cmdsize > @limit name, value = @table.pop @current_table_size -= name.bytesize + value.bytesize + 32 field_arr = @table_by_field[name] field_arr.pop @table_by_field.delete(name) if field_arr.empty? break if @table.empty? end end |
#size_check?(cmdsize) ⇒ Boolean
whether cmd fits in the dynamic table.
To keep the dynamic table size lower than or equal to @limit, remove one or more entries at the end of the dynamic table.
307 308 309 310 |
# File 'lib/http/2/header/encoding_context.rb', line 307 def size_check?(cmdsize) resize_table(cmdsize) cmdsize <= @limit end |
#table_size=(size) ⇒ void
This method returns an undefined value.
Alter dynamic table size. When the size is reduced, some headers might be evicted.
274 275 276 277 |
# File 'lib/http/2/header/encoding_context.rb', line 274 def table_size=(size) @limit = size resize_table(0) end |