Class: HTTP2::Header::EncodingContext
- Inherits:
-
Object
- Object
- HTTP2::Header::EncodingContext
- Includes:
- Error
- Defined in:
- lib/http/2/header/encoding_context.rb
Constant Summary collapse
- UPPER =
/[[:upper:]]/.freeze
- STATIC_TABLE =
[ [":authority", ""], [":method", "GET"], [":method", "POST"], [":path", "/"], [":path", "/index.html"], [":scheme", "http"], [":scheme", "https"], [":status", "200"], [":status", "204"], [":status", "206"], [":status", "304"], [":status", "400"], [":status", "404"], [":status", "500"], ["accept-charset", ""], ["accept-encoding", "gzip, deflate"], ["accept-language", ""], ["accept-ranges", ""], ["accept", ""], ["access-control-allow-origin", ""], ["age", ""], ["allow", ""], ["authorization", ""], ["cache-control", ""], ["content-disposition", ""], ["content-encoding", ""], ["content-language", ""], ["content-length", ""], ["content-location", ""], ["content-range", ""], ["content-type", ""], ["cookie", ""], ["date", ""], ["etag", ""], ["expect", ""], ["expires", ""], ["from", ""], ["host", ""], ["if-match", ""], ["if-modified-since", ""], ["if-none-match", ""], ["if-range", ""], ["if-unmodified-since", ""], ["last-modified", ""], ["link", ""], ["location", ""], ["max-forwards", ""], ["proxy-authenticate", ""], ["proxy-authorization", ""], ["range", ""], ["referer", ""], ["refresh", ""], ["retry-after", ""], ["server", ""], ["set-cookie", ""], ["strict-transport-security", ""], ["transfer-encoding", ""], ["user-agent", ""], ["vary", ""], ["via", ""], ["www-authenticate", ""] ].each(&:freeze).freeze
- STATIC_TABLE_BY_FIELD =
STATIC_TABLE .each_with_object({}) .with_index { |((field, value), hs), idx| (hs[field] ||= []) << [idx, value].freeze } .each_value(&:freeze) .freeze
- STATIC_TABLE_SIZE =
STATIC_TABLE.size
- DEFAULT_OPTIONS =
{ huffman: :shorter, index: :all, table_size: 4096 }.freeze
- STATIC_ALL =
%i[all static].freeze
- STATIC_NEVER =
%i[never static].freeze
Instance Attribute Summary collapse
-
#current_table_size ⇒ Object
readonly
Current table size in octets.
-
#options ⇒ Object
readonly
Current encoding options.
-
#table ⇒ Object
readonly
Current table of header key-value pairs.
Instance Method Summary collapse
-
#addcmd(field, value) ⇒ Hash
Emits command for a header.
-
#dereference(index) ⇒ Array
Finds an entry in current dynamic table by index.
-
#dup ⇒ EncodingContext
Duplicates current compression context.
-
#encode(headers) ⇒ Array
Plan header compression according to @options [:index] :never Do not use dynamic table or static table reference at all.
-
#initialize(options = {}) ⇒ EncodingContext
constructor
Initializes compression context with appropriate client/server defaults and maximum size of the dynamic table.
- #listen_on_table ⇒ Object
-
#process(cmd) ⇒ Array?
Header Block Processing - tools.ietf.org/html/draft-ietf-httpbis-header-compression-10#section-4.1.
-
#table_size=(size) ⇒ Object
Alter dynamic table size.
Constructor Details
#initialize(options = {}) ⇒ EncodingContext
Initializes compression context with appropriate client/server defaults and maximum size of the dynamic table.
119 120 121 122 123 124 125 126 127 |
# File 'lib/http/2/header/encoding_context.rb', line 119 def initialize( = {}) @table = [] @table_by_field = Hash.new { |hs, k| hs[k] = [] } @unshifts = 0 @options = DEFAULT_OPTIONS.merge() @limit = @options[:table_size] @_table_updated = false @current_table_size = 0 end |
Instance Attribute Details
#current_table_size ⇒ Object (readonly)
Current table size in octets
110 111 112 |
# File 'lib/http/2/header/encoding_context.rb', line 110 def current_table_size @current_table_size end |
#options ⇒ Object (readonly)
Current encoding options
:table_size Integer maximum dynamic table size in bytes
:huffman Symbol :always, :never, :shorter
:index Symbol :all, :static, :never
107 108 109 |
# File 'lib/http/2/header/encoding_context.rb', line 107 def @options end |
#table ⇒ Object (readonly)
Current table of header key-value pairs.
100 101 102 |
# File 'lib/http/2/header/encoding_context.rb', line 100 def table @table end |
Instance Method Details
#addcmd(field, value) ⇒ Hash
Emits command for a header. Prefer static table over dynamic table. Prefer exact match over name-only match.
@options [:index] controls whether to use the dynamic table, static table, or both.
:never Do not use dynamic table or static table reference at all.
:static Use static table only.
:all Use all of them.
270 271 272 273 274 275 276 277 278 279 280 281 282 283 284 285 286 287 288 289 290 291 292 293 294 295 296 297 298 299 300 301 302 303 304 305 306 307 |
# File 'lib/http/2/header/encoding_context.rb', line 270 def addcmd(field, value) # @type var exact: Integer? exact = nil # @type var name_only: Integer? name_only = nil index_type = @options[:index] if STATIC_ALL.include?(index_type) && STATIC_TABLE_BY_FIELD.key?(field) STATIC_TABLE_BY_FIELD[field].each do |i, svalue| name_only ||= i if value == svalue exact = i break end end end if index_type == :all && !exact field_entries = @table_by_field[field] field_entries&.each do |hvalue, unshift_id| abs_index = (@unshifts - unshift_id) + STATIC_TABLE_SIZE name_only ||= abs_index if value == hvalue exact = abs_index break end end end if exact { name: exact, type: :indexed } else { name: name_only || field, value: value, type: :incremental } end end |
#dereference(index) ⇒ Array
Finds an entry in current dynamic table by index. Note that index is zero-based in this module.
If the index is greater than the last index in the static table, an entry in the dynamic table is dereferenced.
If the index is greater than the last header index, an error is raised.
156 157 158 159 160 161 162 163 164 165 |
# File 'lib/http/2/header/encoding_context.rb', line 156 def dereference(index) # NOTE: index is zero-based in this module. return STATIC_TABLE[index] if index < STATIC_TABLE_SIZE idx = index - STATIC_TABLE_SIZE raise CompressionError, "Index too large" if idx >= @table.size @table[index - STATIC_TABLE_SIZE] end |
#dup ⇒ EncodingContext
Duplicates current compression context
131 132 133 134 135 136 137 138 139 140 141 142 143 144 |
# File 'lib/http/2/header/encoding_context.rb', line 131 def dup other = EncodingContext.new(@options) t = @table tbf = @table_by_field.transform_values(&:dup) unshifts = @unshifts l = @limit other.instance_eval do @table = t.dup # shallow copy @table_by_field = tbf @unshifts = unshifts @limit = l end other end |
#encode(headers) ⇒ Array
Plan header compression according to @options [:index]
:never Do not use dynamic table or static table reference at all.
:static Use static table only.
:all Use all of them.
241 242 243 244 245 246 247 248 249 250 251 252 253 254 255 |
# File 'lib/http/2/header/encoding_context.rb', line 241 def encode(headers) # Literals commands are marked with :noindex when index is not used noindex = STATIC_NEVER.include?(@options[:index]) headers.each do |field, value| # Literal header names MUST be translated to lowercase before # encoding and transmission. field = field.downcase if UPPER.match?(field) value = "/" if field == ":path" && value.empty? cmd = addcmd(field, value) cmd[:type] = :noindex if noindex && cmd[:type] == :incremental process(cmd) yield cmd end end |
#listen_on_table ⇒ Object
316 317 318 319 320 |
# File 'lib/http/2/header/encoding_context.rb', line 316 def listen_on_table yield ensure @_table_updated = false end |
#process(cmd) ⇒ Array?
Header Block Processing
173 174 175 176 177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 195 196 197 198 199 200 201 202 203 204 205 206 207 208 209 210 211 212 213 214 215 216 217 218 219 220 221 222 223 224 225 226 227 228 229 230 231 232 |
# File 'lib/http/2/header/encoding_context.rb', line 173 def process(cmd) type = cmd[:type] name = cmd[:name] value = cmd[:value] case type when :changetablesize raise CompressionError, "tried to change table size after adding elements to table" if @_table_updated # we can receive multiple table size change commands inside a header frame. However, # we should blow up if we receive another frame where the new table size is bigger. table_size_updated = @limit != @options[:table_size] raise CompressionError, "dynamic table size update exceed limit" if !table_size_updated && value > @limit self.table_size = value nil when :indexed # Indexed Representation # An _indexed representation_ entails the following actions: # o The header field corresponding to the referenced entry in either # the static table or dynamic table is added to the decoded header # list. dereference(name) when :incremental, :noindex, :neverindexed # A _literal representation_ that is _not added_ to the dynamic table # entails the following action: # o The header field is added to the decoded header list. # A _literal representation_ that is _added_ to the dynamic table # entails the following actions: # o The header field is added to the decoded header list. # o The header field is inserted at the beginning of the dynamic table. case name when Integer name, v = dereference(name) value ||= v when UPPER raise ProtocolError, "Invalid uppercase key: #{name}" end emit = [name, value] # add to table if type == :incremental && size_check?(name.bytesize + value.bytesize + 32) @table.unshift(emit) @unshifts += 1 @table_by_field[name].unshift([value, @unshifts]) @current_table_size += name.bytesize + value.bytesize + 32 @_table_updated = true end emit else raise CompressionError, "Invalid type: #{type}" end end |
#table_size=(size) ⇒ Object
Alter dynamic table size.
When the size is reduced, some headers might be evicted.
311 312 313 314 |
# File 'lib/http/2/header/encoding_context.rb', line 311 def table_size=(size) @limit = size resize_table(0) end |