Class: HTTP2::Header::EncodingContext
- Inherits:
-
Object
- Object
- HTTP2::Header::EncodingContext
- Includes:
- Error
- Defined in:
- lib/http/2/compressor.rb
Overview
To decompress header blocks, a decoder only needs to maintain a dynamic table as a decoding context. No other state information is needed.
Constant Summary collapse
- STATIC_TABLE =
[ [':authority', ''], [':method', 'GET'], [':method', 'POST'], [':path', '/'], [':path', '/index.html'], [':scheme', 'http'], [':scheme', 'https'], [':status', '200'], [':status', '204'], [':status', '206'], [':status', '304'], [':status', '400'], [':status', '404'], [':status', '500'], ['accept-charset', ''], ['accept-encoding', 'gzip, deflate'], ['accept-language', ''], ['accept-ranges', ''], ['accept', ''], ['access-control-allow-origin', ''], ['age', ''], ['allow', ''], ['authorization', ''], ['cache-control', ''], ['content-disposition', ''], ['content-encoding', ''], ['content-language', ''], ['content-length', ''], ['content-location', ''], ['content-range', ''], ['content-type', ''], ['cookie', ''], ['date', ''], ['etag', ''], ['expect', ''], ['expires', ''], ['from', ''], ['host', ''], ['if-match', ''], ['if-modified-since', ''], ['if-none-match', ''], ['if-range', ''], ['if-unmodified-since', ''], ['last-modified', ''], ['link', ''], ['location', ''], ['max-forwards', ''], ['proxy-authenticate', ''], ['proxy-authorization', ''], ['range', ''], ['referer', ''], ['refresh', ''], ['retry-after', ''], ['server', ''], ['set-cookie', ''], ['strict-transport-security', ''], ['transfer-encoding', ''], ['user-agent', ''], ['vary', ''], ['via', ''], ['www-authenticate', ''] ].each { |pair| pair.each(&:freeze).freeze }.freeze
Instance Attribute Summary collapse
-
#options ⇒ Object
readonly
Current encoding options.
-
#table ⇒ Object
readonly
Current table of header key-value pairs.
Instance Method Summary collapse
-
#addcmd(*header) ⇒ Hash
Emits command for a header.
-
#current_table_size ⇒ Integer
Returns current table size in octets.
-
#dereference(index) ⇒ Array
Finds an entry in current dynamic table by index.
-
#dup ⇒ EncodingContext
Duplicates current compression context.
-
#encode(headers) ⇒ Array
Plan header compression according to @options [:index] :never Do not use dynamic table or static table reference at all.
-
#initialize(**options) ⇒ EncodingContext
constructor
Initializes compression context with appropriate client/server defaults and maximum size of the dynamic table.
-
#process(cmd) ⇒ Array?
Header Block Processing - tools.ietf.org/html/draft-ietf-httpbis-header-compression-10#section-4.1.
-
#table_size=(size) ⇒ Object
Alter dynamic table size.
Constructor Details
#initialize(**options) ⇒ EncodingContext
Initializes compression context with appropriate client/server defaults and maximum size of the dynamic table.
99 100 101 102 103 104 105 106 107 108 |
# File 'lib/http/2/compressor.rb', line 99 def initialize(**) = { huffman: :shorter, index: :all, table_size: 4096 } @table = [] = .merge() @limit = [:table_size] end |
Instance Attribute Details
#options ⇒ Object (readonly)
Current encoding options
:table_size Integer maximum dynamic table size in bytes
:huffman Symbol :always, :never, :shorter
:index Symbol :all, :static, :never
90 91 92 |
# File 'lib/http/2/compressor.rb', line 90 def end |
#table ⇒ Object (readonly)
Current table of header key-value pairs.
83 84 85 |
# File 'lib/http/2/compressor.rb', line 83 def table @table end |
Instance Method Details
#addcmd(*header) ⇒ Hash
Emits command for a header. Prefer static table over dynamic table. Prefer exact match over name-only match.
@options [:index] controls whether to use the dynamic table, static table, or both.
:never Do not use dynamic table or static table reference at all.
:static Use static table only.
:all Use all of them.
233 234 235 236 237 238 239 240 241 242 243 244 245 246 247 248 249 250 251 252 253 254 255 256 257 258 259 260 261 262 263 264 265 |
# File 'lib/http/2/compressor.rb', line 233 def addcmd(*header) exact = nil name_only = nil if i[all static].include?([:index]) STATIC_TABLE.each_index do |i| if STATIC_TABLE[i] == header exact ||= i break elsif STATIC_TABLE[i].first == header.first name_only ||= i end end end if [:all].include?([:index]) && !exact @table.each_index do |i| if @table[i] == header exact ||= i + STATIC_TABLE.size break elsif @table[i].first == header.first name_only ||= i + STATIC_TABLE.size end end end if exact { name: exact, type: :indexed } elsif name_only { name: name_only, value: header.last, type: :incremental } else { name: header.first, value: header.last, type: :incremental } end end |
#current_table_size ⇒ Integer
Returns current table size in octets
276 277 278 |
# File 'lib/http/2/compressor.rb', line 276 def current_table_size @table.inject(0) { |r, (k, v)| r + k.bytesize + v.bytesize + 32 } end |
#dereference(index) ⇒ Array
Finds an entry in current dynamic table by index. Note that index is zero-based in this module.
If the index is greater than the last index in the static table, an entry in the dynamic table is dereferenced.
If the index is greater than the last header index, an error is raised.
133 134 135 136 137 138 139 |
# File 'lib/http/2/compressor.rb', line 133 def dereference(index) # NOTE: index is zero-based in this module. value = STATIC_TABLE[index] || @table[index - STATIC_TABLE.size] raise CompressionError, 'Index too large' unless value value end |
#dup ⇒ EncodingContext
Duplicates current compression context
112 113 114 115 116 117 118 119 120 121 |
# File 'lib/http/2/compressor.rb', line 112 def dup other = EncodingContext.new() t = @table l = @limit other.instance_eval do @table = t.dup # shallow copy @limit = l end other end |
#encode(headers) ⇒ Array
Plan header compression according to @options [:index]
:never Do not use dynamic table or static table reference at all.
:static Use static table only.
:all Use all of them.
204 205 206 207 208 209 210 211 212 213 214 215 216 217 218 219 |
# File 'lib/http/2/compressor.rb', line 204 def encode(headers) commands = [] # Literals commands are marked with :noindex when index is not used noindex = i[static never].include?([:index]) headers.each do |field, value| # Literal header names MUST be translated to lowercase before # encoding and transmission. field = field.downcase value = '/' if field == ':path' && value.empty? cmd = addcmd(field, value) cmd[:type] = :noindex if noindex && cmd[:type] == :incremental commands << cmd process(cmd) end commands end |
#process(cmd) ⇒ Array?
Header Block Processing
147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 195 |
# File 'lib/http/2/compressor.rb', line 147 def process(cmd) emit = nil case cmd[:type] when :changetablesize raise CompressionError, 'dynamic table size update exceed limit' if cmd[:value] > @limit self.table_size = cmd[:value] when :indexed # Indexed Representation # An _indexed representation_ entails the following actions: # o The header field corresponding to the referenced entry in either # the static table or dynamic table is added to the decoded header # list. idx = cmd[:name] k, v = dereference(idx) emit = [k, v] when :incremental, :noindex, :neverindexed # A _literal representation_ that is _not added_ to the dynamic table # entails the following action: # o The header field is added to the decoded header list. # A _literal representation_ that is _added_ to the dynamic table # entails the following actions: # o The header field is added to the decoded header list. # o The header field is inserted at the beginning of the dynamic table. if cmd[:name].is_a? Integer k, v = dereference(cmd[:name]) cmd = cmd.dup cmd[:index] ||= cmd[:name] cmd[:value] ||= v cmd[:name] = k end emit = [cmd[:name], cmd[:value]] add_to_table(emit) if cmd[:type] == :incremental else raise CompressionError, "Invalid type: #{cmd[:type]}" end emit end |
#table_size=(size) ⇒ Object
Alter dynamic table size.
When the size is reduced, some headers might be evicted.
269 270 271 272 |
# File 'lib/http/2/compressor.rb', line 269 def table_size=(size) @limit = size size_check(nil) end |